/*
  Copyright (c) 2013, 2014 Jouni Sirén
  Copyright (c) 2016 Genome Research Ltd.

  Author: Jouni Sirén <jouni.siren@iki.fi>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include <cmath>
#include <iostream>
#include <stack>

#include "pdltopk.h"

//--------------------------------------------------------------------------

const std::string PDLTopk::EXTENSION = ".topk";
const std::string PDLTopk::MERGE_EXTENSION = ".topk_merge";

PDLTopk::PDLTopk(const RLCSA& _rlcsa, const std::string& base_name, bool compress_sets, bool merge) :
  rlcsa(_rlcsa), status(st_error), supports_merge(merge),
  tree(0), freqs(0), grammar(0), blocks(0)
{
  if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
  {
    std::cerr << "PDLTopk::PDLTopk(): Invalid RLCSA!" << std::endl;
    return;
  }

  std::string input_name = base_name;
  if(compress_sets) { input_name += PDLTree::EXTENSION; }
  else if(this->supportsMerge()) { input_name += MERGE_EXTENSION; }
  else { input_name += EXTENSION; }
  std::ifstream input(input_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "PDLTopk::PDLTopk(): Cannot open input file " << input_name << std::endl;
    return;
  }
  this->tree = new PDLTree(this->rlcsa, input);
  if(!(this->tree->isOk())) { return; }
  this->freqs = CSA::MultiArray::readFrom(input);
  if(!(this->freqs->isOk())) { return; }
  this->status = st_unfinished;

  if(compress_sets)
  {
    uint terminals = 0;
    this->grammar = readGrammar(*(this->tree), base_name, terminals);
    this->blocks = readBlocks(*(this->tree), base_name, terminals);
  }
  else
  {
    pair_type temp;
    input.read((char*)&temp, sizeof(temp)); // Items, bits.
    this->grammar = new CSA::ReadBuffer(input, temp.first, temp.second);
    this->blocks = CSA::MultiArray::readFrom(input);
  }
  input.close();
  if(this->blocks == 0 || !(this->blocks->isOk())) { return; }

  this->status = st_ok;
}

PDLTopk::~PDLTopk()
{
  delete this->tree; this->tree = 0;
  delete this->freqs; this->freqs = 0;
  delete this->grammar; this->grammar = 0;
  delete this->blocks; this->blocks = 0;
}

void
PDLTopk::writeTo(const std::string& base_name) const
{
  if(!(this->isOk()))
  {
    std::cerr << "PDLTopk::writeTo(): Status is not ok!" << std::endl;
    return;
  }

  std::string filename = base_name + (this->supportsMerge() ? MERGE_EXTENSION : EXTENSION);
  std::ofstream output(filename.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "PDLTopk::writeTo(): Cannot open output file " << filename << std::endl;
    return;
  }

  this->tree->writeTo(output);
  this->freqs->writeTo(output);

  pair_type temp(this->grammar->getNumberOfItems(), this->grammar->getItemSize());
  output.write((char*)&temp, sizeof(temp));
  this->grammar->writeTo(output);
  this->blocks->writeTo(output);

  output.close();
}

usint
PDLTopk::reportSize(bool include_freqs) const
{
  usint bytes = sizeof(*this) + this->tree->reportSize() + this->grammar->reportSize() + this->blocks->reportSize();
  if(include_freqs) { bytes += this->freqs->reportSize(); }
  return bytes;
}

//--------------------------------------------------------------------------

PDLTopk::result_type*
PDLTopk::query(const std::string& pattern, usint k, bool get_freqs) const
{
  if(!(this->isOk())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return 0; }
  return this->queryUnsafe(sa_range, k, get_freqs);
}

PDLTopk::result_type*
PDLTopk::query(pair_type sa_range, usint k, bool get_freqs) const
{
  if(!(this->isOk()) || CSA::isEmpty(sa_range) || sa_range.second >= this->rlcsa.getSize()) { return 0; }
  return this->queryUnsafe(sa_range, k, get_freqs);
}

PDLTopk::result_type*
PDLTopk::queryUnsafe(pair_type sa_range, usint k, bool get_freqs) const
{
  usint block = this->tree->getBlock(sa_range);
  if(block >= this->tree->getNumberOfNodes())
  {
    return bruteForceTopk(this->rlcsa, sa_range, k);
  }

  result_type* results = new result_type;
  this->addBlock(results, block, k, get_freqs);
  return results;
}

void
PDLTopk::addBlock(result_type* results, usint block, usint k, bool get_freqs) const
{
  if(k == 0) { k = CSA::WORD_MAX; }

  CSA::MultiArray::Iterator* block_iter = this->blocks->getIterator();
  block_iter->goToItem(block, 0); block_iter->setEnd(block + 1, 0);

  CSA::MultiArray::Iterator* freq_iter = 0;
  pair_type next_freqs(0, 0);
  if(get_freqs)
  {
    freq_iter = this->freqs->getIterator(); freq_iter->goToItem(block, 0);
    next_freqs.first = freq_iter->nextItem();
    next_freqs.second = freq_iter->nextItem();
  }

  std::stack<usint> buffer;
  while(results->size() < k && !(block_iter->atEnd() && buffer.empty()))
  {
    results->push_back(this->next(block_iter, freq_iter, next_freqs, buffer));
  }

  delete block_iter; block_iter = 0;
  delete freq_iter; freq_iter = 0;
}

Document
PDLTopk::next(CSA::MultiArray::Iterator* block_iter, CSA::MultiArray::Iterator* freq_iter,
  pair_type& next_freqs, std::stack<usint>& buffer) const
{
  if(buffer.empty()) { buffer.push(block_iter->nextItem()); }

  while(true)
  {
    usint value = buffer.top(); buffer.pop();
    if(this->tree->isTerminal(value))
    {
      Document doc(value, next_freqs.first);
      if(freq_iter != 0)
      {
        next_freqs.second--;
        if(next_freqs.second == 0)
        {
          if(block_iter->atEnd() && buffer.empty()) { next_freqs = pair_type(0, 0); }
          else
          {
            next_freqs.first = freq_iter->nextItem();
            next_freqs.second = freq_iter->nextItem();
          }
        }
      }
      return doc;
    }
    else
    {
      value = this->tree->toRule(value);
      buffer.push(this->grammar->readItemConst(2 * value + 1));
      buffer.push(this->grammar->readItemConst(2 * value));
    }
  }
}

//--------------------------------------------------------------------------

PDLTopk::result_type*
PDLTopk::query(const std::string& pattern, usint k, merge_type strategy) const
{
  if(!(this->isOk()) || !(this->supportsMerge())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  if(CSA::isEmpty(sa_range)) { return 0; }
  return this->queryUnsafe(sa_range, k, strategy);
}

PDLTopk::result_type*
PDLTopk::query(pair_type sa_range, usint k, merge_type strategy) const
{
  if(!(this->isOk()) || !(this->supportsMerge()) ||
     CSA::isEmpty(sa_range) || sa_range.second >= this->rlcsa.getSize())
  {
    return 0;
  }
  return this->queryUnsafe(sa_range, k, strategy);
}

void
deleteIterators(std::vector<CSA::MultiArray::Iterator*>& iterators)
{
  for(usint i = 0; i < iterators.size(); i++)
  {
    delete iterators[i]; iterators[i] = 0;
  }
}

PDLTopk::result_type*
PDLTopk::queryUnsafe(pair_type sa_range, usint k, merge_type strategy) const
{
  std::vector<usint> blocks;
  this->tree->getBlocks(sa_range, blocks);
  if(blocks.empty())
  {
    return bruteForceTopk(this->rlcsa, sa_range, k);
  }

  result_type* results = new result_type;
  if(strategy == merge_brute)
  {
    for(usint i = 0; i < blocks.size(); i++) { this->addBlock(results, blocks[i], 0, true); }
    sortById(results);
    mergeDocumentRuns(results);
    if(results->size() > k) { results->resize(k); }
    return results;
  }

  scored_type doc_buffer;
  std::vector<CSA::MultiArray::Iterator*> block_iters(blocks.size(), 0), freq_iters(blocks.size(), 0);
  std::vector<pair_type> next_freqs(blocks.size()); // (freq, run length)
  std::vector<std::stack<usint>> buffers;
  std::vector<uint> cumulative_freqs(blocks.size() + 1, 0);
  for(usint i = 0; i < blocks.size(); i++)
  {
    block_iters[i] = this->blocks->getIterator();
    block_iters[i]->goToItem(blocks[i], 0); block_iters[i]->setEnd(blocks[i] + 1, 0);
    freq_iters[i] = this->freqs->getIterator(); freq_iters[i]->goToItem(blocks[i], 0);
    next_freqs[i].first = freq_iters[i]->nextItem();
    next_freqs[i].second = freq_iters[i]->nextItem();
  }

  usint amount = 2 * k;
  while(true)
  {
    // Add more documents from each block.
    if(!(this->addFromBlocks(doc_buffer, amount, block_iters, freq_iters, next_freqs, buffers))) { break; }
    sortById(doc_buffer);
    mergeDocumentRuns(doc_buffer);
    sortByScore(doc_buffer);

    // cumulative_freqs[i] is an upper bound for the total frequency for a document in the remaining
    // blocks, if it has been seen in i blocks.
    for(usint i = 0; i < blocks.size(); i++) { cumulative_freqs[i] = next_freqs[i].first; }
    CSA::sequentialSort(cumulative_freqs.begin(), cumulative_freqs.begin() + blocks.size());
    for(usint i = blocks.size(); i > 0; i--) { cumulative_freqs[i - 1] += cumulative_freqs[i]; }

    // Check whether one of the non-topk documents can still end up in top-k.
    double max_estimate = cumulative_freqs[0];
    for(usint i = k; i < doc_buffer.size(); i++)
    {
      max_estimate = std::max(max_estimate, doc_buffer[i].score + cumulative_freqs[doc_buffer[i].terms]);
    }
    if(doc_buffer.size() >= k && max_estimate <= doc_buffer[k - 1].score) { break; }

    if(strategy == merge_exp) { amount *= 2; }
  }

  deleteIterators(block_iters); deleteIterators(freq_iters);
  for(usint i = 0; i < k && i < doc_buffer.size(); i++) { results->push_back(doc_buffer[i].getDoc()); }
  return results;
}

bool
PDLTopk::addFromBlocks(scored_type& doc_buffer, usint amount,
  std::vector<CSA::MultiArray::Iterator*> block_iters,
  std::vector<CSA::MultiArray::Iterator*> freq_iters,
  std::vector<pair_type>& next_freqs,
  std::vector<std::stack<usint>>& buffers) const
{
  bool found = false;

  for(usint i = 0; i < block_iters.size(); i++)
  {
    if(block_iters[i] == 0 || (block_iters[i]->atEnd() && buffers[i].empty())) { continue; }
    for(usint added = 0; added < amount && !(block_iters[i]->atEnd() && buffers[i].empty()); added++)
    {
      doc_buffer.push_back(ScoredDoc(this->next(block_iters[i], freq_iters[i], next_freqs[i], buffers[i])));
    }
    found = true;
  }

  return found;
}

//--------------------------------------------------------------------------

std::vector<ScoredDoc>*
PDLTopk::query(const std::vector<std::string>& terms, usint k, query_type type, const SadaSparse& counter) const
{
  if(!(this->isOk()) || !(counter.isOk()) || this->supportsMerge() || k == 0) { return 0; }
  if(terms.size() > ScoredDoc::MAX_TERMS)
  {
    std::cerr << "PDLTopk::query(): Number of terms is too large (" << terms.size()
              << ", maximum " << ScoredDoc::MAX_TERMS << ")" << std::endl;
    return 0;
  }

  // Determine the blocks and inverse document frequencies for each term.
  std::vector<ScoredDoc>* results = new std::vector<ScoredDoc>;
  std::vector<CSA::MultiArray::Iterator*> block_iters(terms.size(), 0), freq_iters(terms.size(), 0);
  std::vector<pair_type> next_freqs(terms.size());  // (freq, run length)
  std::vector<double> idfs(terms.size(), 0.0);
  for(usint i = 0; i < terms.size(); i++)
  {
    pair_type range = this->rlcsa.count(terms[i]);
    if(CSA::isEmpty(range))
    {
      if(type == ranked_and)
      {
        deleteIterators(block_iters);
        deleteIterators(freq_iters);
        delete results; results = 0;
        return 0;
      }
      else { continue; }
    }
    idfs[i] = std::log(this->rlcsa.getNumberOfSequences() / (double)(counter.count(range)));
    usint block = this->tree->getBlock(range);
    if(block >= this->tree->getNumberOfNodes()) // Document set not stored for that term.
    {
      result_type* temp = bruteForceTopk(this->rlcsa, range, CSA::WORD_MAX);
      if(temp != 0)
      {
        for(result_type::iterator iter = temp->begin(); iter != temp->end(); ++iter)
        {
          results->push_back(ScoredDoc(*iter, i, idfs[i]));
        }
        delete temp; temp = 0;
      }
    }
    else
    {
      block_iters[i] = this->blocks->getIterator();
      block_iters[i]->goToItem(block, 0); block_iters[i]->setEnd(block + 1, 0);
      freq_iters[i] = this->freqs->getIterator();
      freq_iters[i]->goToItem(block, 0);
      usint freq = freq_iters[i]->nextItem(); usint run = freq_iters[i]->nextItem();
      next_freqs[i] = pair_type(freq, run);
    }
  }

  std::vector<std::stack<usint>> buffers(terms.size());
  usint amount = 2 * k;
  ScoredDoc::mask_type
    all_terms = (~(ScoredDoc::mask_type)0) >> (CHAR_BIT * sizeof(ScoredDoc::mask_type) - terms.size());
  while(true)
  {
    // Add more documents from each block.
    if(!(this->addFromBlocks(*results, amount, block_iters, freq_iters, next_freqs, buffers, idfs))) { break; }
    sortById(*results);
    mergeDocumentRuns(*results);
    ScoredDoc::mask_type finished = 0;
    if(type == ranked_and)  // Remove documents that can no longer contain all terms.
    {
      for(usint i = 0; i < terms.size(); i++)
      {
        if(block_iters[i] == 0 || (block_iters[i]->atEnd() && buffers[i].empty()))
        {
          finished |= ((ScoredDoc::mask_type)1) << i;
        }
      }
      filterByMask(*results, finished);
    }
    sortByScore(*results);

    // Find an upper bound for the score of a document not in the current top-k.
    // For ranked-OR queries, also remove documents that cannot enter top-k any longer.
    double max_estimate = 0.0;
    for(usint i = 0; i < terms.size(); i++) { max_estimate += next_freqs[i].first * idfs[i]; }
    usint tail = std::min(k, (usint)(results->size()));
    for(usint i = k; i < results->size(); i++)
    {
      ScoredDoc& doc = (*results)[i];
      double estimate = doc.score;
      for(usint j = 0; j < terms.size(); j++)
      {
        if(!(doc.terms & (((ScoredDoc::mask_type)1) << j))) { estimate += next_freqs[j].first * idfs[j]; }
      }
      if(type == ranked_or && estimate > (*results)[k - 1].score)
      {
        (*results)[tail] = (*results)[i]; tail++;
      }
      max_estimate = std::max(max_estimate, estimate);
    }
    if(type == ranked_or) { results->resize(tail); }

    // Check that we have found all terms in the top-k documents.
    bool all_found = true;
    if(type == ranked_and)
    {
      usint limit = std::min(k, (usint)(results->size()));
      for(usint i = 0; i < limit; i++)
      {
        if((*results)[i].terms != all_terms) { all_found = false; }
      }
    }

    // FIXME do we want to continue until the first k scores are correct?
    if(results->size() >= k && max_estimate <= (*results)[k - 1].score && all_found) { break; }
    amount = 2 * amount;  // Exponential strategy.
  }

  deleteIterators(block_iters); deleteIterators(freq_iters);
  if(type == ranked_and) { filterByMask(*results, all_terms); }
  if(results->size() > k) { results->resize(k); }
  return results;
}

bool
PDLTopk::addFromBlocks(scored_type& doc_buffer, usint amount,
  std::vector<CSA::MultiArray::Iterator*> block_iters,
  std::vector<CSA::MultiArray::Iterator*> freq_iters,
  std::vector<pair_type>& next_freqs,
  std::vector<std::stack<usint>>& buffers,
  const std::vector<double>& idfs) const
{
  bool found = false;

  for(usint i = 0; i < block_iters.size(); i++)
  {
    if(block_iters[i] == 0 || (block_iters[i]->atEnd() && buffers[i].empty())) { continue; }
    for(usint added = 0; added < amount && !(block_iters[i]->atEnd() && buffers[i].empty()); added++)
    {
      doc_buffer.push_back(ScoredDoc(this->next(block_iters[i], freq_iters[i], next_freqs[i], buffers[i]),
        i, idfs[i]));
    }
    found = true;
  }

  return found;
}

//--------------------------------------------------------------------------
