/*
  Copyright (c) 2013, 2014 Jouni Sirén
  Copyright (c) 2016 Genome Research Ltd.

  Author: Jouni Sirén <jouni.siren@iki.fi>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>
#include <unistd.h>

#include "pdltopk.h"

//--------------------------------------------------------------------------

const static usint DEFAULT_K = 10;

void printUsage();

//--------------------------------------------------------------------------

int
main(int argc, char** argv)
{
  if(argc < 3)
  {
    printUsage();
    std::exit(EXIT_SUCCESS);
  }

  // Parameters
  int c = 0;
  usint k = DEFAULT_K, threads = omp_get_max_threads();
  std::string base_name, query_name;
  PDLTopk::query_type type = PDLTopk::ranked_or;
  while((c = getopt(argc, argv, "ak:ot:")) != -1)
  {
    switch(c)
    {
    case 'a':
      type = PDLTopk::ranked_and; break;
    case 'k':
      k = std::stoul(optarg); break;
    case 'o':
      type = PDLTopk::ranked_or; break;
    case 't':
      threads = CSA::bound(std::stoul(optarg), 1, omp_get_max_threads());
      break;
    case '?':
      std::exit(EXIT_FAILURE);
    default:
      std::exit(EXIT_FAILURE);
    }
  }
  if(optind + 1 >= argc)
  {
    printUsage();
    std::exit(EXIT_SUCCESS);
  }
  base_name = argv[optind];
  query_name = argv[optind + 1];

  std::stringstream ss;
  ss << "Ranked-" << (type == PDLTopk::ranked_and ? "AND" : "OR");
  std::string short_query_type = ss.str();
  ss << "-" << k;
  std::string long_query_type = ss.str();

  std::cout << "TF-IDF query test" << std::endl;
  std::cout << std::endl;
  printHeader("Base name"); std::cout << base_name << std::endl;
  printHeader("Query file"); std::cout << query_name << std::endl;
  printHeader("Query type"); std::cout << short_query_type << std::endl;
  printHeader("k"); std::cout << k << std::endl;
  printHeader("Threads"); std::cout << threads << std::endl;
  std::cout << std::endl;

  // Index structures
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { std::exit(EXIT_FAILURE); }
  PDLTopk topk(rlcsa, base_name, false, false);
  if(!(topk.isOk())) { std::exit(EXIT_FAILURE); }
  SadaSparse counter(rlcsa, base_name);
  if(!(counter.isOk())) { std::exit(EXIT_FAILURE); }

  rlcsa.printInfo();
  usint total_bytes = rlcsa.reportSize() + topk.reportSize(true) + counter.reportSize();
  printSizeNoCSA("RLCSA", rlcsa.getSize(), rlcsa.reportSize());
  printSizeNoCSA("PDL", rlcsa.getSize(), topk.reportSize(true));
  printSizeNoCSA("SadaSparse", rlcsa.getSize(), counter.reportSize());
  printSizeNoCSA("Total", rlcsa.getSize(), total_bytes);
  std::cout << std::endl;

  // Queries
  std::vector<std::string> queries;
  CSA::readRows(query_name, queries, true);
  printHeader("Queries"); std::cout << queries.size() << std::endl;
  std::cout << std::endl;

  omp_set_num_threads(threads);
  double start = CSA::readTimer();
  usint found = 0, docc = 0;
  #pragma omp parallel for schedule(static)
  for(usint i = 0; i < queries.size(); i++)
  {
    std::vector<std::string> terms;
    parseTREC(queries[i], terms);
    std::vector<ScoredDoc>* results = topk.query(terms, k, type, counter);
    if(results != 0)
    {
      if(results->size() > 0)
      {
        #pragma omp atomic
        found++;
      }
      #pragma omp atomic
      docc += results->size();
      delete results; results = 0;
    }
  }
  double seconds = CSA::readTimer() - start;

  printHeader(long_query_type); std::cout << "Found matches for " << found << " queries" << std::endl;
  printResults(long_query_type, seconds, docc, queries.size());
  std::cout << std::endl;

  return 0;
}

//--------------------------------------------------------------------------

void
printUsage()
{
  std::cerr << "Usage: query_tfidf [options] base_name patterns" << std::endl;
  std::cerr << "  -a    Ranked-AND queries" << std::endl;
  std::cerr << "  -o    Ranked-OR queries (default)" << std::endl;
  std::cerr << "  -k N  List top N documents (default " << DEFAULT_K << ")" << std::endl;
  std::cerr << "  -t N  Use N threads (default and max " << omp_get_max_threads() <<  ")" << std::endl;
  std::cerr << std::endl;
}

//--------------------------------------------------------------------------
