/*
  Copyright (c) 2013, 2014 Jouni Sirén
  Copyright (c) 2016 Genome Research Ltd.

  Author: Jouni Sirén <jouni.siren@iki.fi>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#ifndef _DOCLIST_SADACOUNT_H
#define _DOCLIST_SADACOUNT_H

#include <rlcsa/bits/vectors.h>

#include <sdsl/bit_vectors.hpp>

#include "utils.h"

typedef enum { sc_standard, sc_export, sc_filter } sc_mode_type;
void buildSadaCount(const std::string& base_name, sc_mode_type mode, bool print);

extern const std::string SADACOUNT_EXTENSION;    // .sadacount
extern const std::string SADAEXPORT_EXTENSION;   // .sadaexport
extern const std::string SADAFILTER_EXTENSION;   // .sadafilter

//--------------------------------------------------------------------------

template<class V>
class SadaCount
{
  public:
    const static usint BLOCK_SIZE = 32;

    typedef enum { st_unfinished, st_error, st_ok } status_type;

    SadaCount(const RLCSA& _rlcsa, const std::string& base_name) :
      rlcsa(_rlcsa), status(st_error), data(0)
    {
      if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
      {
        std::cerr << "SadaCount::SadaCount(): Invalid RLCSA!" << std::endl;
        return;
      }

      std::string input_name = base_name + SADACOUNT_EXTENSION;
      std::ifstream input(input_name.c_str(), std::ios_base::binary);
      if(!input)
      {
        std::cerr << "SadaCount::SadaCount(): Cannot open input file " << input_name << std::endl;
        return;
      }

      usint bits = 0;
      input.read((char*)&bits, sizeof(bits));
      CSA::ReadBuffer buffer(input, bits, 1);
      this->data = CSA::encode<V>(buffer, BLOCK_SIZE, bits);

      input.close();
      this->status = st_ok;
    }

    ~SadaCount()
    {
      delete this->data; this->data = 0;
    }

    inline bool isOk() const { return (this->status == st_ok); }

    void writeTo(const std::string& base_name) const
    {
      if(!(this->isOk()))
      {
        std::cerr << "SadaCount::writeTo(): Status is not ok!" << std::endl;
        return;
      }

      std::string filename = base_name + SADACOUNT_EXTENSION;
      std::ofstream output(filename.c_str(), std::ios_base::binary);
      if(!output)
      {
        std::cerr << "SadaCount::writeTo(): Cannot open output file " << filename << std::endl;
        return;
      }

      usint bits = this->data->getSize();
      output.write((char*)&bits, sizeof(bits));
      CSA::ReadBuffer* buffer = CSA::decode<V>(*(this->data));
      if(buffer != 0)
      {
        buffer->writeTo(output);
        delete buffer; buffer = 0;
      }

      output.close();
    }

    usint reportSize() const
    {
      usint bytes = sizeof(*this);

      if(this->data != 0) { bytes += this->data->reportSize(); }

      return bytes;
    }

    usint count(const std::string& pattern) const
    {
      if(!(this->isOk())) { return 0; }
      pair_type sa_range = this->rlcsa.count(pattern);
      return this->count(sa_range);
    }

    usint count(pair_type sa_range) const
    {
      if(!(this->isOk()) || sa_range.second >= this->rlcsa.getSize()) { return 0; }

      usint len = CSA::length(sa_range);
      if(len <= 1) { return len; }

      typename V::Iterator iter(*(this->data));
      return 2 * len - (iter.select(sa_range.second) - iter.select(sa_range.first)) - 1;
    }

  private:
    const RLCSA&     rlcsa;
    status_type      status;

    // Unlike in the original paper, we encode h(v) as an 1-bit, followed by h(v) 0-bits.
    V*               data;

    // These are not allowed.
    SadaCount();
    SadaCount(const SadaCount&);
    SadaCount& operator = (const SadaCount&);
};

//--------------------------------------------------------------------------

template<class V, class F>
class SadaFilter
{
  public:
    const static usint BLOCK_SIZE = 32;

    typedef enum { st_unfinished, st_error, st_ok } status_type;

    SadaFilter(const RLCSA& _rlcsa, const std::string& base_name) :
      rlcsa(_rlcsa), status(st_error), data(0), filter(0)
    {
      if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
      {
        std::cerr << "SadaFilter::SadaFilter(): Invalid RLCSA!" << std::endl;
        return;
      }

      std::string input_name = base_name + SADAFILTER_EXTENSION;
      std::ifstream input(input_name.c_str(), std::ios_base::binary);
      if(!input)
      {
        std::cerr << "SadaFilter::SadaFilter(): Cannot open input file " << input_name << std::endl;
        return;
      }

      { // Vector
        usint bits = 0;
        input.read((char*)&bits, sizeof(bits));
        CSA::ReadBuffer buffer(input, bits, 1);
        this->data = CSA::encode<V>(buffer, BLOCK_SIZE, bits);
      }
      { // Filter
        usint bits = 0;
        input.read((char*)&bits, sizeof(bits));
        CSA::ReadBuffer buffer(input, bits, 1);
        this->filter = CSA::encode<F>(buffer, BLOCK_SIZE, bits);
      }

      input.close();
      this->status = st_ok;
    }

    ~SadaFilter()
    {
      delete this->data; this->data = 0;
      delete this->filter; this->filter = 0;
    }

    inline bool isOk() const { return (this->status == st_ok); }

    void writeTo(const std::string& base_name) const
    {
      if(!(this->isOk()))
      {
        std::cerr << "SadaFilter::writeTo(): Status is not ok!" << std::endl;
        return;
      }

      std::string filename = base_name + SADAFILTER_EXTENSION;
      std::ofstream output(filename.c_str(), std::ios_base::binary);
      if(!output)
      {
        std::cerr << "SadaFilter::writeTo(): Cannot open output file " << filename << std::endl;
        return;
      }

      { // Vector
        usint bits = this->data->getSize();
        output.write((char*)&bits, sizeof(bits));
        CSA::ReadBuffer* buffer = CSA::decode<V>(*(this->data));
        if(buffer != 0)
        {
          buffer->writeTo(output);
          delete buffer; buffer = 0;
        }
      }
      { // Filter
        usint bits = this->filter->getSize();
        output.write((char*)&bits, sizeof(bits));
        CSA::ReadBuffer* buffer = CSA::decode<F>(*(this->filter));
        if(buffer != 0)
        {
          buffer->writeTo(output);
          delete buffer; buffer = 0;
        }
      }

      output.close();
    }

    usint reportSize() const
    {
      usint bytes = sizeof(*this);

      if(this->data != 0) { bytes += this->data->reportSize(); }
      if(this->filter != 0) { bytes += this->filter->reportSize(); }

      return bytes;
    }

    usint count(const std::string& pattern) const
    {
      if(!(this->isOk())) { return 0; }
      pair_type sa_range = this->rlcsa.count(pattern);
      return this->count(sa_range);
    }

    usint count(pair_type sa_range) const
    {
      if(!(this->isOk()) || sa_range.second >= this->rlcsa.getSize()) { return 0; }

      usint len = CSA::length(sa_range);
      if(len <= 1) { return len; }

      sa_range.second--; // Convert to a range of internal nodes.
      typename F::Iterator f_iter(*(this->filter));
      if(sa_range.first > 0) { sa_range.first -= f_iter.rank(sa_range.first - 1); }
      sa_range.second -= f_iter.rank(sa_range.second);
      if(CSA::isEmpty(sa_range)) { return 1; } // Only a filtered subtree remaining.

      sa_range.second++; // Back to SA range.
      len = CSA::length(sa_range);

      typename V::Iterator iter(*(this->data));
      return 2 * len - (iter.select(sa_range.second) - iter.select(sa_range.first)) - 1;
    }

  private:
    const RLCSA&     rlcsa;
    status_type      status;

    // Unlike in the original paper, we encode h(v) as an 1-bit, followed by h(v) 0-bits.
    V*               data;
    F*               filter;

    // These are not allowed.
    SadaFilter();
    SadaFilter(const SadaFilter&);
    SadaFilter& operator = (const SadaFilter&);
};

//--------------------------------------------------------------------------

/*
  Reimplementation of Sada-S.
*/

class SadaSparse
{
  public:
    typedef enum { st_unfinished, st_error, st_ok } status_type;
    typedef sdsl::sd_vector<>                       sd_vector;

    SadaSparse(const RLCSA& _rlcsa, const std::string& base_name);
    ~SadaSparse();

    inline bool isOk() const { return (this->status == st_ok); }

    void writeTo(const std::string& base_name) const;
    usint reportSize() const;

    usint count(const std::string& pattern) const;
    usint count(pair_type sa_range) const;

  private:
    const RLCSA&             rlcsa;
    status_type              status;

    // Positions with non-zero values are marked with an 1-bit.
    sd_vector                filter;
    sd_vector::rank_1_type   filter_rank;

    // Non-zero values encoded in unary: k becomes 0^{k-1} 1.
    sd_vector                values;
    sd_vector::select_1_type value_select;

    // These are not allowed.
    SadaSparse();
    SadaSparse(const SadaSparse&);
    SadaSparse& operator = (const SadaSparse&);
};

//--------------------------------------------------------------------------

#endif  // _DOCLIST_SADACOUNT_H
